import { apc, modelName, modelSymbol, parentSymbol, vpc } from './symbols';
import mitt from './emitter';
var objKeys = Object.keys.bind(Object);
function diff(oldObj, newObj, whitelist) {
    if (oldObj === newObj)
        return;
    objKeys(oldObj)
        .reduce(function (acc, val) {
        return ~acc.indexOf(val) ? acc : acc.concat(val);
    }, objKeys(newObj))
        .forEach(function (key) {
        if (~whitelist.indexOf(key))
            return;
        var oldVal = oldObj[key];
        var newVal = newObj[key];
        if (oldVal === newVal)
            return;
        if (oldVal != null &&
            newVal != null &&
            typeof oldVal === 'object' &&
            typeof newVal === 'object') {
            oldVal[modelSymbol]
                ? oldVal.applySnapshot(newVal, true)
                : diff(oldVal, newVal, whitelist);
        }
        else {
            oldObj[apc][key] = newVal;
        }
    });
}
function setUpObject(obj, emit, symbol, path, parent) {
    if (parent === void 0) { parent = obj; }
    if (obj[modelName] && !obj[modelSymbol]) {
        obj = modelMap.has(obj[modelName])
            ? modelMap.get(obj[modelName])(obj)
            : obj;
    }
    if (obj[modelSymbol] && obj !== parent) {
        obj[parentSymbol](emit, parent, path);
        return obj;
    }
    else {
        // have to ignore errors because typescript doesn't like indexing with symbols
        if (!obj[apc] ||
            // @ts-ignore
            !obj[apc][symbol])
            obj[apc] = cProxy(obj, emit, symbol, false, path, parent);
        if (!obj[vpc] ||
            // @ts-ignore
            !obj[vpc][symbol])
            obj[vpc] = cProxy(obj, emit, symbol, true, path, parent);
    }
    for (var prop in obj) {
        var val = obj[prop];
        if (val != null && typeof val === 'object') {
            obj[prop] = setUpObject(val, emit, symbol, path + '/' + prop, parent);
        }
    }
    return obj;
}
function cProxy(obj, emit, symbol, view, path, parent) {
    return new Proxy(obj, {
        get: function (target, prop) {
            if (prop === symbol)
                return true;
            if (typeof prop === 'symbol')
                return target[prop];
            var res = target[prop];
            // check if there are other built in constructors that need this
            if (target instanceof Date && typeof res === 'function')
                return res.bind(target);
            return ((res != null &&
                typeof res === 'object' &&
                (view ? res[vpc] : res[apc])) ||
                res);
        },
        set: function (target, prop, val) {
            if (view) {
                throw new Error("You can't modify the state outside of actions");
            }
            if (target[prop] === val)
                return true;
            if (typeof prop === 'symbol') {
                target[prop] = val;
                return true;
            }
            if (val != null && typeof val === 'object') {
                val = setUpObject(val, emit, symbol, path + '/' + prop, parent || target[vpc]);
            }
            target[prop] = val;
            emit('patch', { path: path + '/' + prop, op: 'replace', value: val });
            return true;
        },
    });
}
function subscribe(emitter, evt, fn) {
    emitter.on(evt, fn);
    return function () { return emitter.off(evt, fn); };
}
var modelMap = new Map();
export default function model(name, init) {
    function instantiate(obj) {
        if (obj === void 0) { obj = {}; }
        var symbol = Symbol('model');
        var path = '';
        var parentEmit = null;
        var parent = null;
        var emitter = mitt();
        var model = init(obj);
        var emit = function (evt, val) {
            if (evt === 'snapshot') {
                val = common.getSnapshot();
            }
            else if (val.path != null) {
                val.path = path + val.path;
            }
            emitter.emit(evt, val);
            parentEmit && parentEmit(evt, val);
        };
        var whitelist = [];
        var state = {};
        var common = (_a = {
                onAction: function (fn, after) {
                    var evt = after ? 'action-complete' : 'action';
                    return subscribe(emitter, evt, fn);
                },
                onSnapshot: function (fn) {
                    return subscribe(emitter, 'snapshot', fn);
                },
                onPatch: function (fn) {
                    return subscribe(emitter, 'patch', fn);
                },
                getSnapshot: function () {
                    return JSON.parse(JSON.stringify(state));
                },
                applySnapshot: function (snapshot, dontemit) {
                    diff(state, snapshot, whitelist);
                    !dontemit && emit('snapshot', {});
                },
                getParent: function () {
                    return parent;
                },
                getRoot: function () {
                    return parent ? parent.getRoot() : state;
                }
            },
            _a[modelSymbol] = true,
            _a[parentSymbol] = function (emit, _parent, _path) {
                parentEmit = emit;
                path = _path;
                parent = _parent;
            },
            _a[modelName] = name,
            _a);
        whitelist.push.apply(whitelist, objKeys(common));
        state = setUpObject(common, emit, symbol, '');
        // r = refresh on get, v = value
        var caches = {};
        var emitSnapshot = function (name, args) {
            emit('action-complete', { name: name, path: '', args: args });
            emit('snapshot', {});
        };
        var _loop_1 = function (key) {
            var prop = Object.getOwnPropertyDescriptor(model, key);
            if (!prop)
                return "continue";
            if (typeof prop.get === 'function') {
                // view
                var view_1 = prop.get;
                Object.defineProperty(state, key, {
                    get: function () {
                        var cache = caches[key];
                        if (cache.r) {
                            cache.r = false;
                            cache.v = view_1.apply(state[vpc]);
                        }
                        return cache.v;
                    },
                    configurable: false,
                });
                caches[key] = { v: undefined, r: true };
            }
            else if (typeof prop.value === 'function') {
                // action
                var action_1 = prop.value;
                state[key] = function () {
                    var args = [];
                    for (var _i = 0; _i < arguments.length; _i++) {
                        args[_i] = arguments[_i];
                    }
                    emit('action', { name: key, path: '', args: args });
                    var res = action_1.apply(state[apc], args);
                    if (res != null && res.then) {
                        res.then(emitSnapshot.bind(null, key, args));
                    }
                    else {
                        emitSnapshot(key, args);
                    }
                    return res;
                };
                whitelist.push(key);
            }
            else {
                // prop
                state[apc][key] = prop.value;
            }
        };
        for (var key in model) {
            _loop_1(key);
        }
        // update the cache on change
        state.onPatch(function () {
            for (var key in caches) {
                caches[key].r = true;
            }
        });
        return state;
        var _a;
    }
    modelMap.set(name, instantiate);
    return function (obj) { return instantiate(obj)[vpc]; };
}
export function clearCache() {
    modelMap.clear();
}
//# sourceMappingURL=index_new.js.map